include(unixas.h)
        TITLE({ibtrtn - fortran bit primitives {08-feb-82}})

	SUBTITLE({integer function  ibtand(a,b)})
	PROC(ibtand,,{{a},{b}})
	mcoml	*a,r0			# .not.a
	bicl3	r0,*b,r0		# .not.(.not.a) .and. b
	ret


	SUBTITLE({integer function  ibtcom(a)})
	PROC(ibtcom,,{{a}})
	mcoml	*a,r0			# .not.a
	ret


	SUBTITLE({logical function  ibtest(string,n)})
#
# The article 
#
# Nelson  H.F.  Beebe,   "Proposal  for a Standard Set of Primitives for
# Machine-Independent Bit Manipulation in FORTRAN", Proceedings  of  the
# NRCC  Conference   on  Software Standards in Chemistry, 1979, Lawrence
# Berkeley Laboratory, Berkeley, CA
#
# describes the bit string primitives IBTEST,  IBTON,  and  IBTOFF,  and
# states that bits must be numbered from 1..N starting with the LEFTMOST
# bit in a word.  For string applications entirely in memory, the  order
# is  not significant, but for the construction of strings to be  output
# to a device, the order is  critical.   Raster  devices  receive  8-bit
# bytes  numbered  from  high  to low as 1..8, and successive bytes  are
# concatenated on the right to form the bit pattern defining the raster.
#
# The bit numbering in the VAX does not conform  to  this  model,  since
# bits are numbered from right to left 0..31, and bytes are  transmitted
# in the order 0 (7..0), 1 (15..8), 2 (23..16), 3 (31..24), 4 (7..0),  5
# (15..8),  etc.   We therefore chose to number bits in BYTE order,  but
# with reversed numbering within the byte.  In this way,  the  user  can
# picture  a  string  of  consecutive bytes in memory at byte offsets  0
# (1..8), 1 (9..16), 2 (17..24), 3 (25..32), 4 (33..40), 5  (41..48),  6
# (49..56), 7 (57..64), etc.  The relation between our numbering and VAX
# order is
#
# ouroffset = our bit # - 1
#
# VAX bit # = (ouroffset/8)*8 + 7 - mod(ouroffset,8)
#           = (ouroffset/8)*8 + 7 - {ouroffset - (ouroffset/8)*8}
#           = 7 - ouroffset + (ouroffset/8)*16
#
# Our ordering will therefore conform to the proposal if the  "word"  is
# interpreted  to  be "8-bit byte", which is reasonable since a  byte is
# the smallest immediately addressable chunk of memory.
#
# Since the offset is non-negative, two arithmetic shifts can be used to
# generate the last term of the above expression.
#
	PROC(ibtest,,{{string},{n}})
	movl	*n,r0			# n
	bleq	testfalse		# exit if n <= 0
	decl	r0			# n-1 = offset in string
	ashl	$-3,r0,r1		# (offset)/8
	ashl	$4,r1,r1		# 16*(offset/8)
	addl2	$7,r1			# 7 + 16*(offset/8)
	subl3	r0,r1,r0		# 7 + 16*(offset/8) - offset=vax offset
	extzv	r0,$1,*string,r0	# get bit number n in r0
	ret				# bit = 1 ==> .true., 
testfalse:				# bit = 0 ==> .false.
	clrl	r0
	ret


	SUBTITLE({subroutine  ibtoff(string,n)})
	PROC(ibtoff,,{{string},{n}})
	movl	*n,r0			# n
	bleq	offdon			# exit if n <= 0
	decl	r0			# n-1 = offset in bit string
	ashl	$-3,r0,r1		# (offset)/8
	ashl	$4,r1,r1		# 16*(offset/8)
	addl2	$7,r1			# 7 + 16*(offset/8)
	subl3	r0,r1,r0		# 7 + 16*(offset/8) - offset=vax offset
	insv	$0,r0,$1,*string	# turn off bit n
offdon:	ret


	SUBTITLE({subroutine  ibton(string,n)})

	PROC(ibton,,{{string},{n}})
	movl	*n,r0			# n
	bleq	ondone			# exit if n <= 0
	decl	r0			# n-1 = offset in bit string
	ashl	$-3,r0,r1		# (offset)/8
	ashl	$4,r1,r1		# 16*(offset/8)
	addl2	$7,r1			# 7 + 16*(offset/8)
	subl3	r0,r1,r0		# 7 + 16*(offset/8) - offset=vax offset
	insv	$1,r0,$1,*string	# turn on bit n
ondone:	ret


	SUBTITLE({integer function  ibtor(a,b)})
	PROC(ibtor,,{{a},{b}})
	bisl3	*a,*b,r0
	ret


	SUBTITLE({integer function  ibtrol(a,n)})
	PROC(ibtrol,{2,3,4},{{a},{n}})
	movl	*a,r0			# a
	movl	*n,r3			# n
	bgeq	rolpos
	mnegl	r3,r3
rolpos:	beql	roldon			# exit if n = 0
	decl	r3			# n-1
	emul	r3,$1,$0,r0		# (n-1) into (r0,r1)
	ediv	$32,r0,r0,r3		# mod(n-1,32) into r3
	incl	r3			# mod(n-1,32)+1 = value in 1..32
	subl3	r3,$32,r3		# cr=32-abs(n) = count for a right shift
	extzv	$0,r3,*a,r1		# get bits 0..cr-1 into r1
	subl3	r3,$32,r4		# cl=32-cr
	extzv	r3,r4,*a,r2		# get bits cl..32 into r2
	insv	r1,r4,r3,r0		# put 0..cr-1 into cl..32
	insv	r2,$0,r4,r0		# put cl..32 into cr..32
roldon:	ret



	SUBTITLE({integer function  ibtror(a,n)})
	PROC(ibtror,{2,3,4},{{a},{n}})
	movl	*a,r0			# a
	movl	*n,r3			# n
	bgeq	rorpos
	mnegl	r3,r3
rorpos:	beql	rordon			# exit if n = 0
	decl	r3			# n-1
	emul	r3,$1,$0,r0		# (n-1) into (r0,r1)
	ediv	$32,r0,r0,r3		# mod(n-1,32) into r3
	incl	r3			# mod(n-1,32)+1 = value in 1..32
	extzv	$0,r3,*a,r1		# get bits 0..cr-1 into r1
	subl3	r3,$32,r4		# cl=32-cr
	extzv	r3,r4,*a,r2		# get bits cl..32 into r2
	insv	r1,r4,r3,r0		# put 0..cr-1 into cl..32
	insv	r2,$0,r4,r0		# put cl..32 into cr..32
rordon:	ret


	SUBTITLE({integer function  ibtrot(a,n)})
	PROC(ibtrot,,{{a},{n}})
	.globl	_ibtrol_
	.globl	_ibtror_
	pushl	*n			# put arguments back on stack
	pushl	*a
	tstl	*n
	blss	rotright
	calls	$2,_ibtrol_
	ret
rotright:
	calls	$2,_ibtror_
	ret


	SUBTITLE({integer function  ibtshf(a,n)})
	PROC(ibtshf,,{{a},{n}})
	.globl	_ibtshl_
	.globl	_ibtshr_
	pushl	*n			# put arguments back on stack
	pushl	*a
	tstl	*n
	blss	shfright
	calls	$2,_ibtshl_
	ret
shfright:
	calls	$2,_ibtshr_
	ret


	SUBTITLE({integer function  ibtshl(a,n)})
	PROC(ibtshl,,{{a},{n}})
	movl	*n,r0
	bgeq	shlpos
	mnegl	r0,r0
shlpos:	ashl	r0,*a,r0		# left shift a into r0
	ret


	SUBTITLE({integer function  ibtshr(a,n)})
	PROC(ibtshr,{2},{{a},{n}})
	movl	*n,r0
	blss	shlneg
	mnegl	r0,r0
shlneg:	clrl	r2			# -abs(n) in r0
	movl	*a,r1			# (a,0) in (r1,r2)
	ashq	r0,r1,r0		# right shift (r1,r2) into (r0,r1)
	ret


	SUBTITLE({integer function  ibtsum(a)})
	PROC(ibtsum,{2},{{a}})
	clrl	r0			# bit count
	movl	$31,r2			# loop limit (31..0 by -1)
	movl	*a,r1			# a

sumloop:
	bbc	r2,r1,nextbit		# jump if bit = 0 at position r2
	incl	r0			# high bit = 1, count it
nextbit:sobgeq	r2,sumloop		# loop 32 times

	ret				# ibtsum = r0
	

	SUBTITLE({integer function  ibtxor(a,b)})
	PROC(ibtxor,,{{a},{b}})
	xorl3	*a,*b,r0
	ret

	END	
